#!/usr/bin/env python
# -*- coding: utf-8 -*-

import ConfigParser
import logging
import os

_logger = logging.getLogger(__name__)


class ConfigManager(object):
    """!
    Config Manager.
    Prüft, ob die Dateien für die Ofen-Konfiguration vorliegen ansonsten werden
    die default-Konfigurationen geladen
    """
    def __init__(self, pathfile=None):
        """!
        Konstruktor
        Initialisiert den Config-Manager
        Prüft übergebenen Pfad auf Config-File (hier stehen die Pfade für verschiedene Verzeichnisse/Dateien
        und speichert die Pfade in einer Liste
        @param pathfile: das zu suchende bzw. zu durchsuchende Config-File
        """
        _logger.debug("ConfigManager init")
        self._path_file_parser = None
        self._config_parser = ConfigParser.ConfigParser()
        self._configfile_path_available = False
        self._production_data_parser = ConfigParser.ConfigParser()
        self._production_data_parser_available = False
        self._parser = ConfigParser.ConfigParser()

        path_list = []  # a list with all path to configuration files like *.gs1, *.cnf

        if os.path.isfile(pathfile):
            path_to_pathfile = pathfile
            _logger.debug("found configuration files ...")
        else:
            path_to_pathfile = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'part_r', 'pathfile.gs1')
            _logger.error("use default configuration files ... ")

        try:
            self._path_file_parser = ConfigParser.ConfigParser()
            self._path_file_parser.read(path_to_pathfile)
            path_list.append(path_to_pathfile)
        except IOError:
            _logger.fatal("pathfile.gs1 not found")

        self.reload(path_list)

    def reload(self, path_list=None):
        """!
        Reload der Konfigurations-Files
        configfile.gs1, produktionsdaten.gs1, default.cnf
        @param path_list: Liste mit den Pfaden zu den Konfigurations-Files
        """

        if path_list is None:
            path_list = []

        # configfile.gs1
        try:
            configfile_path = self._path_file_parser.get('path', 'configfile')
            self._config_parser.read(configfile_path)
            path_list.append(configfile_path)
            self._configfile_path_available = True
            # _logger.debug("configfile.gs1 loaded ...")
        except IOError:
            _logger.critical("configfile.gs1 not found")
            self._configfile_path_available = False
            exit()

        # produktionsdaten.gs1
        try:
            production_path = self._path_file_parser.get('path', 'produktionsdaten')
            self._production_data_parser.read(production_path)
            path_list.append(production_path)
            self._production_data_parser_available = True
            # _logger.debug("produktionsdaten.gs1 loaded ...")
        except IOError:
            _logger.critical("produktionsdaten.gs1 not found")
            self._production_data_parser_available = False
            exit()

        # all config-files together
        try:
            self._parser.read(path_list)
            # _logger.debug("configuration ready")
        except:     # TODO: Exception Type
            _logger.error("error parser initialisation")

    def get_configfile_parser(self):
        """!
        Gibt den Parser für das file configfile.gs1 zurück
        @return Parser für das file configfile.gs1
        """
        # _logger.debug("get configfile.gs1")
        if self._configfile_path_available:
            return self._config_parser
        else:
            _logger.critical("configfile.gs1 not available on system")
            return None

    def get_parser(self):
        """!
        Gibt den Parser für alle config-Einträge zurück
        @return Parser für alle config-Einträge
        """
        return self._parser

    def get_path_file_parser(self):
        """!
        Gibt den Parser für die Pfad-Datei zurück
        @return Parser für die Pfad-Datei
        """
        return self._path_file_parser

    def get_config_file_value(self, heading, key, default='0'):
        """!
        Liest aus dem configfile.gs1 File
        @param heading: Name der Section im Ini-File
        @param key: Name des Keys im Ini-File
        @param default: Rückgabewert, falls die Datei oder der Parameter nicht gelesen werden kann
        @return Wert des angegeben heading/key-Paares oder Default-Wert
        """
        try:
            return self._config_parser.get(heading, key)
        except ConfigParser.Error:
            _logger.error("can't read {}, {} from configfile.gs1. Using default = {}".format(heading, key, default))
            return default

    def get_production_file_value(self, heading, key, default='0'):
        """!
        Liest aus dem produktionsdaten.gs1 File
        @param heading: Name der Section im Ini-File
        @param key: Name des Keys im Ini-File
        @param default: Rückgabewert, falls die Datei oder der Parameter nicht gelesen werden kann
        @return Wert des angegeben heading/key-Paares oder Default-Wert
        """
        try:
            return self._production_data_parser.get(heading, key)
        except ConfigParser.Error:
            _logger.error("can't read {}, {} from produktionsdaten.gs1. Using default = {}".format(heading, key, default))
            return default

    def set_production_file_value(self, heading, key, value):
        """!
        Schreibt direkt in das produktionsdaten.gs1 File
        @param heading: Name der Section im Ini-File
        @param key: Name des Keys im Ini-File
        @param value: Value des Keys im Ini-File
        """
        try:
            default_path = self._path_file_parser.get('path', 'produktionsdaten')
            self._production_data_parser.set(heading, key, value)
            with open(default_path, 'wb') as configfile:
                self._production_data_parser.write(configfile)
        except ConfigParser.Error:
            _logger.error("can't write to produktionsdaten.gs1")

    def remove_production_file_value(self, heading, key):
        """!
        Entfernt einen Eintrag im produktionsdaten.gs1 File
        @param heading: Name der Section im Ini-File
        @param key: Name des Keys im Ini-File
        """
        try:
            default_path = self._path_file_parser.get('path', 'produktionsdaten')
            self._production_data_parser.remove_option(heading, key)
            with open(default_path, 'wb') as configfile:
                self._production_data_parser.write(configfile)
        except ConfigParser.Error:
            _logger.error("can't delete key in produktionsdaten.gs1")

    def get_calibration_file_value(self, heading, key, default='0'):
        """!
        Liest aus dem calibration_data.gs1 File
        @param heading: Name der Section im Ini-File
        @param key: Name des Keys im Ini-File
        @param default: Rückgabewert, falls die Datei oder der Parameter nicht gelesen werden kann
        @return Wert des angegeben heading/key-Paares oder Default-Wert
        """
        try:
            default_path = self._path_file_parser.get('path', 'calibration_data')
            default_parser = ConfigParser.ConfigParser()
            default_parser.read(default_path)
            return default_parser.get(heading, key)
        except ConfigParser.Error:
            _logger.error("can't read {}, {} from calibration_data.gs1. Using default = {}".format(heading, key, default))
            return default

    def set_calibration_file_value(self, heading, key, value):
        """!
        Schreibt direkt in das calibration_data.gs1 File
        @param heading: Name der Section im Ini-File
        @param key: Name des Keys im Ini-File
        @param value: Value des Keys im Ini-File
        """
        try:
            default_path = self._path_file_parser.get('path', 'calibration_data')
            default_parser = ConfigParser.ConfigParser()
            default_parser.read(default_path)
            if not default_parser.has_section(heading):
                default_parser.add_section(heading)
            default_parser.set(heading, key, value)
            with open(default_path, 'wb') as configfile:
                default_parser.write(configfile)
        except ConfigParser.Error:
            _logger.error("can't write to calibration_data.gs1")

    def remove_calibration_file_value(self, heading, key):
        """!
        Entfernt einen Eintrag im calibration_data.gs1 File
        @param heading: Name der Section im Ini-File
        @param key: Name des Keys im Ini-File
        """
        try:
            default_path = self._path_file_parser.get('path', 'calibration_data')
            default_parser = ConfigParser.ConfigParser()
            default_parser.read(default_path)
            default_parser.remove_option(heading, key)
            with open(default_path, 'wb') as configfile:
                default_parser.write(configfile)
        except ConfigParser.Error:
            _logger.error("can't delete key in calibration_data.gs1")
