#!/usr/bin/env python
# -*- coding: utf-8 -*-

import Queue
import time
import logging
import collections
from threading import Event, Thread

_logger = logging.getLogger(__name__)
ENABLE_DEBUG_LOG = False

DOOR_CLOSE = 0
DOOR_OPEN = 100

DoorMoveState = collections.namedtuple('DoorMoveState', 'auf zu idle')._make(range(3))


class DoorControl(object):
    """!
    Diese Klasse ist für die Steuerung der Türe zuständig.
    Sie ist für die Überwachung der Endschalter sowie der
    Ansteuerung des Lift-Motors verantwortlich.
    """
    def __init__(self, daio, config_manager, error_handler):
        """!
        Initialisiert die Türsteuerung
        @param daio: Instanz der DIO
        @param analog_input: Instanz der AIO
        @param config_manager: Instanz des Config Managers
        @param error_handler: Instanz des Fehlermanagers
        """
        _logger.debug("DoorControl init")
        self._name = "DoorControl"

        self._daio = daio
        self._error_handler = error_handler

        self._door_open_close_time_configured = int(config_manager.get_config_file_value('door', 'door-closing-time')) / 1000.0    # laut configfile.gs1 ist door-closing-time = 22000 msec
        self._door_follow_up_time = int(config_manager.get_config_file_value('door', 'door-follow-up-time')) / 1000.0              # Nachlaufzeit für Prüfung der Tür-Isolation
        self._door_gap_time = int(config_manager.get_config_file_value('door', 'door-gap-time', 1000)) / 1000.0                     # Zeit, die benötigt wird, damit eine 3mm Lücke zum Brennraum geschlossen wird
        self._door_open_close_time = self._door_open_close_time_configured

        if config_manager.get_config_file_value('door', 'door-variant') == 'electrical':
            self._door_variation = 1  # 1 means 'electrical' (config_file 30)
        elif config_manager.get_config_file_value('door', 'door-variant') == 'mechanical':
            self._door_variation = 2  # 2 means 'mechanical' (config_file 30)
        else:
            _logger.error("config value not found")

        self._output_buffer = Queue.Queue()
        self._door_error_active = False
        self._actual_door_running_time = 0
        self._door_move_time = 0
        self._door_move = DoorMoveState.idle
        self._door_move_to_end_stop = False
        self._door_closed_endstop_reached = False
        self._door_insulation_missing = True
        self._door_self_test = False
        self._door_self_test_success = False
        self._door_state = "STOPPED"

        # Überwachungs Thread einrichten und starten
        self._stopped = Event()
        self._stopped.clear()
        self._process_time = 0.1

        self._doorcontrol_thread = Thread(target=self._process, name=self._name)
        self._doorcontrol_thread.setDaemon(True)

        self._doorcontrol_thread.start()
        _logger.debug("DoorControl thread started")

    def exit(self):
        """!
        Destruktor beendet Thread
        """
        # Überwachungstask beenden
        self._stopped.set()
        time.sleep(self._process_time)

        _logger.debug("DoorControl del")

    def door_open(self, percent=100):
        """!
        Öffnet die Tür zu einem gewissen Prozent-Teil
        @param percent: Gibt an, wie weit die Tür geöffnet werden soll (% des gesamten Weges)
                        Default: Ganz öffnen, da maximale Laufzeit angenommen wird.
                                 Sollte die Türe schon etwas geöffnet sein, stoppt sie im Endschalter.
        @return True, wenn die Türe wie angefordert geöffnet werden kann.
        """
        # _logger.debug(str(percent))
        if 0 <= percent <= 100:                                                             # Bereichsbegrenzung auf [0..100]%
            # if self._door_state == "STOPPED":                                               # nur ausführen, wenn Türe still steht
                if percent == 100:                                                          # Für komplettes Öffnen
                    self._door_move_time = self._door_open_close_time * 1.1                 # Toleranz von 10%
                    self._door_move = DoorMoveState.auf                                     # Bewegungsrichtung: auf
                elif percent == 0:                                                          # Bei 0 wird nichts gemacht
                    self._door_move_time = 0                                                #
                    self._door_move = DoorMoveState.idle                                    #
                else:
                    self._door_move_time = self._door_open_close_time / 100.0 * percent     # Berechnung der Schrittweite (Tür-Position als Laufzeit dargestellt)
                    self._door_move = DoorMoveState.auf                                     # Bewegungsrichtung: auf
                return True
            # else:
            #     _logger.warning("door_open not executed - door is moving")
            #     return False

        else:
            _logger.warning("no valid value - only [0-100] allowed")
            return False

    def door_close(self):
        """!
        Schließt die Tür.
        @return True, wenn Türe geschlossen werden kann. False, wenn Türe nicht geschlossen werden kann.
        """
        if (self._door_state == "STOPPED") or \
           (self._door_state == "OPENING") or \
           (self._door_state == "CLOSING"):                                     # nur, wenn Türe steht, öffnet oder schließt
            self._door_move = DoorMoveState.zu                                  # Bewegungsrichtung: zu
            return True
        else:
            _logger.warning("door_close not executed")
            return False

    @property
    def actual_percent_door(self):
        """!
        Getter für die aktuelle Stellung der Tür in % zurück
        @return Türstellung: 0% - Türe vollständig geschlossen, 100% - Türe vollständig geöffnet
        """
        percent_door = min(int(self._actual_door_running_time / self._door_open_close_time * 100), 100)
        percent_door = max(percent_door, 0)
        return percent_door

    @property
    def door_error(self):
        """!
        Getter für den aktuellen Fehlerzustand der Tür zurück (kein Fehler/Fehler)
        @return True - Fehler in der Türsteuerung, False - Türsteuerung ist Fehlerfrei
        """
        return self._door_error_active

    @property
    def door_self_test(self):
        """!
        Status des Tür-Selbsttests.
        @return Wird ausgeführt (True) oder ist beendet (False).
        """
        return self._door_self_test

    def start_door_self_test(self):
        """!
        Startet den Selbsttest für die Türsteuerung.
        """
        self._door_error_active = False
        self._door_self_test_success = False    # Wegen Neustart, Selbsttest auf nicht bestanden
        self._door_self_test = True

    @property
    def door_self_test_success(self):
        """!
        Ergebnis des Tür-Selbsttests.
        @return True - Selbsttest erfolgreich, False - Fehler während des Selbsttests
        """
        return self._door_self_test_success

    def set_mains_frequency(self, frequency=50):
        """!
        Funktion zur Korrektur der Türlaufzeiten. Die Tür kann je nach Netzfrequenz schneller oder langsamer laufen.
        @param frequency: Von der Leistungseinheit gemessene Netzfrequenz [Hz]. Gültige Werte liegen zwischen 40 und 70 Hz
        """
        if 40 <= frequency <= 70:
            # Korrektur mit Netzfrequenz, da Motor schneller oder langsamer laufen kann
            self._door_open_close_time = self._door_open_close_time_configured / 50 * frequency
            self._door_follow_up_time = self._door_follow_up_time / 50 * frequency
            self._door_gap_time = self._door_gap_time / 50 * frequency
            _logger.debug("door-closing-time correction from {}s to {}s".format(self._door_open_close_time_configured,
                                                                                self._door_open_close_time))
        else:
            _logger.warning("mains frequency out of range: {} Hz".format(frequency))

    def is_door_closed(self):
        """!
        Zustandsabfrage, ob Türe geschlossen ist.
        @return True - Türe ist geschlossen, False - Türe ist nicht geschlossen.
        """
        if self._daio.door_end_switch_close_is_set() or (self._actual_door_running_time < 0):
            return True
        else:
            return False

    def is_door_open(self):
        """!
        Zustandsabfrage, ob Türe offen ist.
        @return True - Türe ist offen, False - Türe ist nicht geöffnet.
        """
        return self._daio.door_end_switch_open_is_set()

    def get_door_state(self):
        """!
        Zustandsabfrage der Türe.
        @return Zustand der Türe.
        """
        if self._door_state == "STOPPED" and self.is_door_open():
            return "OPEN"
        elif self._door_state == "STOPPED" and self.is_door_closed():
             return "CLOSED"
        elif self._door_state == "OPENING":
            return "OPENING"
        elif self._door_state == "CLOSING":
            return "CLOSING"
        elif self._door_state == "CLOSE_GAP":
            return "CLOSE_GAP"
        else:
            return "SELFTEST"

    def is_door_insulation_missing(self):
        """!
        Ergebnis der Prüfung der Türisolation
        @return True - Isolation ist eingesetzt, False - Isolation ist nicht eingesetzt
        """
        return self._door_insulation_missing

    def read(self):
        """!
        Mit dieser Funktion holt der Gs1Controller Meldungen der Türsteuerung an die GUI ab.
        @return Meldungen der Türsteuerung an die GUI. Wenn None, liegt keine Meldung vor
        """
        if self._output_buffer.empty():
            return None
        else:
            rc = self._output_buffer.get()
            self._output_buffer.task_done()
            return rc

    def _process(self):
        """!
        Statemachine Thread für die Türsteuerung.
        """
        _logger.debug("{} process started".format(self._name))

        last_time = time.time()
        timeout_door = 0

        # Ausgangssituation der Tür feststellen
        if self._daio.door_end_switch_open_is_set():                            # Tür ist offen?
            self._actual_door_running_time = self._door_open_close_time         # maximale Laufzeit setzen
        elif self._daio.door_end_switch_close_is_set():                         # Tür ist zu?
            self._actual_door_running_time = 0                                  # Laufzeit auf 0
        else:                                                                   # steht irgendwo
            self._actual_door_running_time = 0                                  # Türe als geschlossen annehmen

        while not self._stopped.wait(self._process_time):
            process_start_time = time.time()
            # Elektrische Türe
            if self._door_variation == 1:

                if self._door_state == "SELFTEST_START":                            # Starten des Selbsttests veranlassen
                    self._door_self_test_success = False                            # Init: Selbsttest auf nicht bestanden
                    self._door_error_active = False                                 # Annahme: Es liegt kein Türfehler vor
                    self._door_move = DoorMoveState.idle                            # Keine Bewegungsrichtung
                    last_time = time.time()                                         # Misst die Zeit, wann die Türe mit der Bewegung startet
                    if self._daio.door_end_switch_close_is_set():                   # Endschalter ist geschlossen?
                        self._door_closed_endstop_reached = True                    # Merker setzen
                        self._door_insulation_missing = False                       # Prüfung der Türisolation vorbereiten -> Annahme, dass diese eingebaut ist.
                        self._actual_door_running_time = 0                          # Laufzeitkorrektur vornehmen
                        timeout_door = self._door_follow_up_time                    # Nachlaufzeit für Prüfung der Tür-Isolation
                    else:
                        self._door_closed_endstop_reached = False                   # Endschalter (geschlossen) Zustand verwerfen
                        self._actual_door_running_time = self._door_open_close_time  # Laufzeit initialisieren
                        timeout_door = self._door_open_close_time * 1.1             # Timeout = Soll-Zeit - 10% Gesamtlaufzeit -> Türe schließt immer komplett!
                    self._daio.lift_zu()                                            # Nach oben Fahren
                    self._door_state = "SELFTEST_CLOSING"                           # neuer Zustand: ST Türe öffnet
                    _logger.debug("SELFTEST_START - closing door")

                elif self._door_state == "SELFTEST_CLOSING":                        # Erster Testschritt: Tür muss sich schließen lassen
                    delta_time = time.time() - last_time                            # Wie lange hat sich die Türe seit dem letzten Aufruf bewegt?
                    self._actual_door_running_time -= delta_time                    # Aktuelle Position der Türe berechnen (Zeit wird kleiner)
                    timeout_door -= delta_time                                      # Timeout wird kleiner
                    last_time = time.time()                                         # Zeitmessung für nächsten Durchlauf

                    if self._door_closed_endstop_reached:                           # Wurde der Endschalter schon einmal erreicht?
                        if not self._daio.door_end_switch_close_is_set():           # Endschalter öffnet wieder?
                            self._door_insulation_missing = True                    # Isolation fehlt!
                            self._output_buffer.put("insulationmissing")            # Meldung an GUI veranlassen
                            _logger.debug("SELFTEST_CLOSING - door insulation is missing")
                            self._daio.lift_auf()                                   # Erster Schritt beendet, Tür öffnen
                            timeout_door = self._door_open_close_time * 1.1         # Timeout = Soll-Zeit - 10% Gesamtlaufzeit -> Türe öffnet immer komplett!
                            self._actual_door_running_time = 0                      # Laufzeit korrigieren
                            self._door_state = "SELFTEST_OPENING"                   # 2. Testschritt: Tür öffnen
                    else:
                        if self._daio.door_end_switch_close_is_set():               # Endschalter wird geschlossen?
                            self._door_closed_endstop_reached = True                # Merker setzen
                            self._door_insulation_missing = False                   # Prüfung der Türisolation vorbereiten -> Annahme, dass diese eingebaut ist.
                            self._actual_door_running_time = 0                      # Laufzeitkorrektur vornehmen
                            timeout_door = self._door_follow_up_time                # Nachlaufzeit für Prüfung der Tür-Isolation
                            _logger.debug("SELFTEST_CLOSING - door close end stop reached")

                    if timeout_door <= 0:                                           # Timeout ist abgelaufen, Bewegung wird beendet.
                        if not self._daio.door_end_switch_close_is_set():           # Tür ist nicht im Endschalter?
                            if not self._door_closed_endstop_reached:               # Er wurde auch nie erreicht?
                                _logger.debug("SELFTEST_CLOSING - door closed end stop not reached")
                                self._door_self_test = False                        # Selbsttest beendet
                                self._daio.lift_stop()                              # Stoppen
                                self._door_state = "ERROR"                          # Türe ist blockiert -> Fehler
                                self._error_handler.set_error("F60")
                            else:
                                self._door_insulation_missing = True                # Isolation fehlt!
                                self._output_buffer.put("insulationmissing")        # Meldung an GUI veranlassen
                                _logger.debug("SELFTEST_CLOSING - door insulation is missing")
                                self._daio.lift_auf()                               # Nach unten Fahren
                                timeout_door = self._door_open_close_time * 1.1     # Timeout = Soll-Zeit - 10% Gesamtlaufzeit -> Türe öffnet immer komplett!
                                self._actual_door_running_time = 0                  # Laufzeit korrigieren, Türe ist zu
                                self._door_state = "SELFTEST_OPENING"               # In Stop Zustand springen
                        else:
                            self._daio.lift_auf()                                   # Nach unten Fahren
                            _logger.debug("SELFTEST_CLOSING - opening door")
                            timeout_door = self._door_open_close_time * 1.1         # Timeout = Soll-Zeit - 10% Gesamtlaufzeit -> Türe öffnet immer komplett!
                            self._door_state = "SELFTEST_OPENING"                   # In Stop Zustand springen

                elif self._door_state == "SELFTEST_OPENING":                        # Zweiter Testschritt: Tür muss sich öffnen lassen
                    delta_time = time.time() - last_time                            # Wie lange hat sich die Türe seit dem letzten Aufruf bewegt?
                    self._actual_door_running_time += delta_time                    # Aktuelle Position der Türe berechnen (Zeit wird größer)
                    timeout_door -= delta_time                                      # Timeout wird kleiner
                    last_time = time.time()                                         # Zeitmessung für nächsten Durchlauf

                    if self._daio.door_end_switch_open_is_set():                    # Endschalter erreicht? (Prio 1)
                        self._daio.lift_stop()                                      # Stoppen
                        self._door_self_test = False                                # Selbsttest beendet
                        self._actual_door_running_time = self._door_open_close_time     # Laufzeit korrigieren
                        _logger.debug("SELFTEST_OPENING - selftest passed")
                        self._door_self_test_success = True                         # Selbsttest bestanden
                        self._error_handler.reset_error("F60")                      # Türfehler zurücksetzen
                        self._error_handler.reset_error("F61")                      # Türfehler zurücksetzen
                        self._error_handler.reset_error("F62")                      # Türfehler zurücksetzen
                        self._output_buffer.put("dooropen")                         # Türstatus mitteilen
                        self._door_state = "STOPPED"                                # sonst passt alles -> Stoppen
                    elif timeout_door <= 0:                                         # Timeout ist abgelaufen, Bewegung wird beendet.
                        self._daio.lift_stop()                                      # Stoppen
                        self._door_self_test = False                                # Selbsttest beendet
                        _logger.debug("SELFTEST_OPENING - door open end stop not reached")
                        self._door_state = "ERROR"                                  # In Fehler Zustand springen
                        self._error_handler.set_error("F60")

                # Zustand Türe gestoppt
                elif self._door_state == "STOPPED":                                 # Zustand: Türe bewegt sich nicht
                    if self._door_move == DoorMoveState.auf:                        # Neue Bewegungsrichtung: auf?
                        self._daio.lift_auf()                                       # Nach unten Fahren
                        last_time = time.time()                                     # Misst die Zeit, wann die Türe mit der Bewegung startet
                        timeout_door = self._door_move_time                         # Timeout = Dauer der gewünschten Bewegungsdauer
                        self._door_closed_endstop_reached = False                   # Endschalter (geschlossen) Zustand verwerfen
                        self._door_state = "OPENING"                                # neuer Zustand: Türe öffnet
                        self._output_buffer.put("dooropen")                         # Teile der GUI die aktuelle Bewegungsrichtung mit
                        _logger.debug("opening door")
                    elif self._door_move == DoorMoveState.zu:                       # Neue Bewegungsrichtung: zu?
                        self._daio.lift_zu()                                        # Nach oben Fahren
                        last_time = time.time()                                     # Misst die Zeit, wann die Türe mit der Bewegung startet
                        timeout_door = self._door_open_close_time * 1.1             # Timeout = Soll-Zeit + 10% Gesamtlaufzeit -> Türe schließt immer komplett!
                        self._door_closed_endstop_reached = False                   # Endschalter (geschlossen) Zustand verwerfen
                        self._door_state = "CLOSING"                                # neuer Zustand: Türe schließt
                        _logger.debug("closing door")
                        self._output_buffer.put("doorclose")                        # Teile der GUI die aktuelle Bewegungsrichtung mit
                    elif self._door_self_test:                                      # Soll der Selbsttest der Türe wiederholt / gestartet werden?
                        self._door_state = "SELFTEST_START"                         # In den Selbsttest wechseln
                    else:
                        self._daio.lift_stop()                                      # Stoppen

                # Zustand Türe öffnet
                elif self._door_state == "OPENING":                                 # Zustand: Türe öffnet
                    delta_time = time.time() - last_time                            # Wie lange hat sich die Türe seit dem letzten Aufruf bewegt?
                    self._actual_door_running_time += delta_time                    # Aktuelle Position der Türe berechnen (Zeit wird größer)
                    timeout_door -= delta_time                                      # Timeout wird kleiner
                    last_time = time.time()                                         # Zeitmessung für nächsten Durchlauf                                         
                    
                    if self._door_move == DoorMoveState.zu:                         # Neue Bewegungsrichtung: zu? (Prio 1)
                        self._daio.lift_stop()                                      # Stoppen
                        self._door_state = "STOPPED"                                # In Stop Zustand springen
                    elif self._daio.door_end_switch_open_is_set():                  # Endschalter erreicht? (Prio 2)
                        self._daio.lift_stop()                                      # Stoppen
                        self._door_move = DoorMoveState.idle                        # Keine Bewegungsrichtung
                        self._actual_door_running_time = self._door_open_close_time     # Laufzeit korrigieren
                        self._door_state = "STOPPED"                                # sonst passt alles -> Stoppen
                    elif self._actual_door_running_time > self._door_open_close_time * 1.05:         # Türe bewegt sich schon zu lange -> Fehler (Prio 3)
                        # TODO: Im realen Betrieb testen, da dieser Fehler im Emulator nicht erkannt wird.
                        self._daio.lift_stop()                                      # Stoppen
                        self._door_move = DoorMoveState.idle                        # Keine Bewegungsrichtung
                        _logger.debug("door open end stop not reached")
                        self._door_state = "ERROR"                                  # In Fehler Zustand springen
                        self._error_handler.set_error("F60")
                    elif timeout_door <= 0:                                         # Die Türe ist soweit offen, wie gewünscht (Prio 4)
                        self._daio.lift_stop()                                      # Stoppen
                        self._door_move = DoorMoveState.idle                        # Keine Bewegungsrichtung
                        _logger.debug("timeout - door stopped")
                        self._door_state = "STOPPED"                                # In Stop Zustand springen

                # Zustand Türe schließt
                elif self._door_state == "CLOSING":                                 # Zustand Türe schließt
                    delta_time = time.time() - last_time                            # Wie lange hat sich die Türe seit dem letzten Aufruf bewegt?
                    self._actual_door_running_time -= delta_time                    # Aktuelle Position der Türe berechnen (Zeit wird kleiner)
                    timeout_door -= delta_time                                      # Timeout wird kleiner
                    last_time = time.time()                                         # Zeitmessung für nächsten Durchlauf

                    if self._door_move == DoorMoveState.auf:                        # Neue Bewegungsrichtung: auf?
                        self._daio.lift_stop()                                      # Stoppen
                        self._door_state = "STOPPED"                                # In Stop Zustand springen
                    elif self._door_closed_endstop_reached:                           # Wurde der Endschalter schon einmal erreicht?
                        if not self._daio.door_end_switch_close_is_set():           # Endschalter öffnet wieder?
                            self._door_insulation_missing = True                    # Isolation fehlt!
                            self._output_buffer.put("insulationmissing")            # Meldung an GUI veranlassen
                            _logger.debug("door insulation is missing")
                            self._daio.lift_stop()                                  # Stoppen
                            self._door_move = DoorMoveState.idle                    # Keine Bewegungsrichtung
                            self._door_state = "STOPPED"                            # In Stop Zustand springen
                            self._output_buffer.put("doorclose")                    # Teile der GUI die aktuelle Bewegungsrichtung mit
                    else:
                        if self._daio.door_end_switch_close_is_set():               # Endschalter wird geschlossen?
                            self._door_closed_endstop_reached = True                # Merker setzen
                            self._output_buffer.put("doorclosed")                   # Meldung an GUI veranlassen
                            self._door_insulation_missing = False                   # Prüfung der Türisolation vorbereiten -> Annahme, dass diese eingebaut ist.
                            self._actual_door_running_time = 0                      # Laufzeitkorrektur vornehmen
                            timeout_door = self._door_follow_up_time                # Nachlaufzeit für Prüfung der Tür-Isolation
                            _logger.debug("door close end stop reached")

                    if timeout_door <= 0:                                           # Timeout ist abgelaufen, Bewegung wird beendet.
                        self._daio.lift_stop()                                      # Stoppen
                        self._door_move = DoorMoveState.idle                        # Keine Bewegungsrichtung

                        if not self._daio.door_end_switch_close_is_set():           # Tür ist nicht im Endschalter?
                            if not self._door_closed_endstop_reached:               # Er wurde auch nie erreicht?
                                _logger.debug("door close end stop not reached")
                                self._door_state = "ERROR"                          # Türe ist blockiert -> Fehler
                                self._error_handler.set_error("F60")
                            else:
                                self._door_insulation_missing = True                # Isolation fehlt!
                                self._output_buffer.put("insulationmissing")        # Meldung an GUI veranlassen
                                _logger.debug("door insulation is missing")
                                self._output_buffer.put("doorclose")                # Teile der GUI die aktuelle Bewegungsrichtung mit
                                self._door_state = "STOPPED"                        # In Stop Zustand springen
                        else:
                            self._output_buffer.put("doorclose")                    # Teile der GUI die aktuelle Bewegungsrichtung mit
                            timeout_door = self._door_gap_time + 0.2                # Zeit, die die Tür zum Schließen der Lücke benötigt
                            self._door_state = "CLOSE_GAP"                          # In Stop Zustand springen

                elif self._door_state == "CLOSE_GAP":
                    delta_time = time.time() - last_time                            # Wie lange hat sich die Türe seit dem letzten Aufruf bewegt?
                    timeout_door -= delta_time                                      # Timeout wird kleiner
                    last_time = time.time()                                         # Zeitmessung für nächsten Durchlauf

                    if timeout_door <= 0:                                           # Timeout ist abgelaufen, Bewegung wird beendet.
                        self._daio.lift_stop()  # Stoppen
                        self._door_move = DoorMoveState.idle  # Keine Bewegungsrichtung
                        self._door_state = "STOPPED"  # In Stop Zustand springen
                    elif timeout_door <= self._door_gap_time:
                        if not(self._door_move == DoorMoveState.zu):                # Bewegt sich die Tür schon?
                            self._door_move = DoorMoveState.zu                      # Türe nochmal zu machen
                            self._daio.lift_zu()                                    # Nach oben Fahren

                # Zustand Tür-Fehler
                elif self._door_state == "ERROR":
                    self._daio.lift_stop()                                          # Stoppen
                    self._door_move = DoorMoveState.idle                            # Keine Bewegungsrichtung
                    if not self._door_error_active:                                 # Wenn noch nicht in Fehlerstate gewesen
                        _logger.debug("door error state")
                        self._door_error_active = True                              # und Fehlerflag setzen

                    if self._door_self_test:                                        # Soll der Selbsttest der Türe wiederholt / gestartet werden?
                        self._door_state = "SELFTEST_START"                         # In den Selbsttest wechseln
                        self._door_error_active = False                             # und Fehlerflag setzen

                # Nirvana-Zustand führt zu Tür-Fehler
                else:
                    self._door_state = "ERROR"                                      # In Fehler Zustand springen
                    self._error_handler.set_error("F60")
                    _logger.debug("door nirvana state")
                    self._daio.lift_stop()                                          # Stoppen
                    self._door_move = DoorMoveState.idle                            # Keine Bewegungsrichtung

                # Schalterüberwachung: Türe muss sich mindestens 30% vom jeweiligen Schalter wegbewegt haben
                if (self._actual_door_running_time < (self._door_open_close_time * 0.7) and
                   self._daio.door_end_switch_open_is_set()):                       # Hängt der untere Endschalter?
                    if self._door_state != "ERROR":
                        _logger.debug("door open end stop error")
                        self._door_state = "ERROR"                                  # Türe ist blockiert -> Fehler
                        self._error_handler.set_error("F61")

                if (self._actual_door_running_time > (self._door_open_close_time * 0.3) and
                   self._daio.door_end_switch_close_is_set()):                      # Endschalter Tür geschlossen hängt?
                    if self._door_state != "ERROR":
                        _logger.debug("door closed end stop error")
                        self._door_state = "ERROR"                                  # In Fehler Zustand springen
                        self._error_handler.set_error("F62")

                if ENABLE_DEBUG_LOG:
                    _logger.debug("{}, {}, {}, {}, {}, {}, {}, {}, {}, {}".format(
                        self._door_error_active,
                        self._actual_door_running_time,
                        self._door_move_time,
                        self._door_move,
                        self._door_move_to_end_stop,
                        self._door_closed_endstop_reached,
                        self._door_insulation_missing,
                        self._door_self_test,
                        self._door_self_test_success,
                        self._door_state))

            # Mechanische Türe
            elif self._door_variation == 2:
                # TODO: Mechanische Türe implementieren. Release 2

                if self._door_self_test:
                    self._door_self_test_success = False  # last self test was okay
                    self._door_self_test = False  # clear self test flag

            delta_process_time = time.time() - process_start_time
            wait_time = self._process_time - delta_process_time
            if wait_time <= 0:
                if __debug__:
                    _logger.warning("no wait time:{}".format(delta_process_time))

        _logger.debug("{} process stopped".format(self._name))
